# Verison 1.2
# Add Read_ESM function used to read STIFFNESS and MASS matrices from files.
# Stop obtaining Element Elastic Energy from ABAQUS, Calculate Objective by external function 
# Delete Resubmit Function for it's useless. The reason, which is not sure enough, why ABAQUS collapse is MemoryError
# Adopt SIMP model which means OC algorithm and sensitivity filter were introduced in this version
# Add some help functions: Modify INP, Pre_Matrix, Result_Plot_SIMP
#Version 1.3
# SIMP Bi-Mateiral Model Static Optimization
# Add Dynamic Optimization
# Modify Obj_Cal for complex calculation
# Fix a bug for Read_ESM function. It can't read MASS matrix before.
# Fix a bug in preFlt function. distance of two elements should be sqrt of the sum of the squares of coordinates.
# Add Self-define excitation frequency of load
# Modify OC algorithm for Non-Negative in sqrt function
#Version 1.4
# Add Power-flow Objective function and redesign code structure which enable the code more compatible for different Objective function.
# Fix excitation frequency. Should be natural(radian) frequency NOT cycles frequency
# Add MMA Algorithm
# Add Structural Intensity Function
# Fix Structural Intensity Function under Global Coordinate System
#Version 1.5
# Add Restore Mechanism
# Add Post-process 
# Revise the programming structure. Add global parameters setting
# Add Final Run
# Revise restore mechanism by global variables

import os
import time
import pickle
import numpy as np
import numpy.linalg
from numpy.ctypeslib import load_library,ndpointer
from ctypes import c_int,c_double
from caeModules import *
import math, customKernel
from abaqusConstants import *
from odbAccess import openOdb

def Read_ESM(ESM_Name, STIF, MASS, NODENUM, N):
	f=open(ESM_Name,'r')
	E_line='Start'
	E_NUM=0    
	while E_line<>'':
		E_line = f.readline()
		if E_line[3:17]=='ELEMENT NUMBER':
			E_NUM=E_NUM+1
		elif E_line[3:16]=='ELEMENT NODES':
			E_line=f.readline();
			NODENUM[E_NUM]=np.array(E_line[2:-1].strip().split(','),dtype=int)
		Head=E_line[8:len(E_line)-1].strip()
		if Head=='TYPE=STIFFNESS' and E_NUM<>0:
			STIF[E_NUM],MASS[E_NUM]=Read_Matrices(f,N,Head[5:len(Head)],E_NUM)
	f.close()
    
def Read_Matrices(f,N,Type,Enum):
	Line_Matrix=[];K=np.zeros([N,N]);M=np.zeros([N,N])
	data_line = f.readline()
	data_line=data_line.strip().split(',');data_line.pop() 
	while data_line[0]<>'**':
		Line_Matrix.extend(data_line)
		data_line = f.readline()
		if data_line=='':break
		data_line=data_line.strip().split(',')
		if data_line[-1]=='': data_line.pop()  
	Line_Matrix.reverse()
	for i in range(1,1+N):    
		for j in range(1,1+i):
			K[i-1,j-1]=Line_Matrix.pop()
			if i<>j:K[j-1,i-1]=K[i-1,j-1]
	Line_Matrix.pop();Line_Matrix.pop()
	for i in range(1,1+N):    
		for j in range(1,1+i):
			M[i-1,j-1]=Line_Matrix.pop()
			if i<>j:M[j-1,i-1]=M[i-1,j-1]
	return K,M

def Dumping_Data_Abaqus():
    U_data={};U_Coordinates={};Ele_Conn={}
    odb = session.openOdb('Initial_Data_Retrieving.odb')
    frame=odb.steps['Step-1'].frames[-1]
    U=frame.fieldOutputs['U'].values
    MyInstance_Elements=odb.rootAssembly.instances['PART-1-1'].elements
    MyInstance_nodes=odb.rootAssembly.instances['PART-1-1'].nodes
    for i in range(len(MyInstance_Elements)): # Loop all elements
        Ele_Conn[i+1]=MyInstance_Elements[i].connectivity
    for i in range(len(U)): # Loop all nodes
        U_data[i+1]=U[i].data
        U_Coordinates[i+1]=MyInstance_nodes[i].coordinates
    # Dump Data
    f=open('Ele_Conn.txt','w')
    pickle.dump(Ele_Conn,f)
    f.close()
    f=open('Nod_Coor.txt','w')
    pickle.dump(U_Coordinates,f)
    f.close()
    f=open('U_data.txt','w')
    pickle.dump(U_data,f)
    f.close()
    
def Modify_INP(INP_name):
    # Open *.INP
    try:
        f=open(INP_name+'.inp','r+')
    except ValueError:
        print 'Can\'t open {0}.inp'.format(INP_name)        
    E_line='Start'
    while E_line<>'':
        E_line = f.readline()
        E_line=E_line.strip()
        if E_line=='*End Step':
            f.seek(-11,2)
            f.write("*FILE FORMAT, ASCII\n*ELEMENT MATRIX OUTPUT, ELSET=LAYER-1.whole, FILE NAME=ESMLAYER3600-2, STIFFNESS=YES, MASS=YES, OUTPUT FILE=USER DEFINED\n*ELEMENT MATRIX OUTPUT, ELSET=BASE-1.whole, FILE NAME=ESMBASE3600, STIFFNESS=YES, MASS=YES, OUTPUT FILE=USER DEFINED\n*End Step\n")
            E_line=''
    f.close()

def preFlt(Rmin,Elmts,Nds):
	Fm={}
	# Calculate element centre coordinates
	elm, c0 = np.zeros(len(Elmts)), np.zeros((len(Elmts),3))
	for i in range(len(elm)):
		elm[i] = Elmts[i].label
		nds = Elmts[i].connectivity
		for nd in nds: c0[i] = np.add(c0[i],np.divide(Nds[nd].coordinates,len(nds)))
	# Weighting factors
	for i in range(len(elm)):
		Fm[elm[i]] = [[],[]]
		for j in range(len(elm)):
			dis = np.sqrt(np.sum(np.power(np.subtract(c0[i],c0[j]),2)))
			if dis<Rmin:
				Fm[elm[i]][0].append(elm[j])
				Fm[elm[i]][1].append(Rmin - dis)
		Fm[elm[i]][1] = np.divide(Fm[elm[i]][1],np.sum(Fm[elm[i]][1]))
	return Fm

def fltAe(Ae,Fm,Xe):
	raw = Ae.copy()
	for el in Fm.keys():
		Ae[el-1] = 0.0
		for i in range(len(Fm[el][0])): Ae[el-1]+=raw[Fm[el][0][i]-1]*Fm[el][1][i]*Xe[Fm[el][0][i]-1]/Xe[el-1]
	return Ae

def BESO(Vf,Xe,Ae,Ele_NUM):
	lo, hi = min(Ae), max(Ae)
	tv =Vf*Ele_NUM
	while (hi-lo)/hi > 1.0e-5:
		th = (lo+hi)/2.0
		for i in range(Ele_NUM): 
			if Ae[i]>th:
				Xe[i] = 1.0
			else:
				Xe[i] = 0.001
		if np.sum(Xe)-tv>0: lo = th
		else: hi = th
	# Label elements as solid or void
	slb, vlb = [], []
	for i in range(Ele_NUM):
		if Xe[i] == 1.0: 
			slb.append(i+1)
		else: 
			vlb.append(i+1)
	return Xe,slb,vlb

def OC(Vf,Xe,Ae,Ele_NUM):
	lo, hi, move = 0, 1e9, 0.1
	tv = Vf*Ele_NUM
	while (hi-lo)/(hi+lo) > 1.0e-3:
		lmid = (lo+hi)/2.0
		xnew=np.maximum(0.001,np.maximum(Xe-move,np.minimum(1,np.minimum(Xe+move,Xe*np.sqrt(np.maximum(0,-Ae/lmid))))))
		if sum(xnew)-tv>0: 
			lo = lmid 
		else: 
			hi = lmid
	return xnew

def FE_Mdb(mdb,E,D,N,w,alpha,beta,th,possoin):
	# Params Define
	mdl = mdb.models['Model-1']
	part = mdl.parts[Part_Name]
	#for i in range(len(mdl.materials)):
	#	del mdl.materials[mdl.materials.keys()[0]]
	#for i in range(len(part.sectionAssignments)):
	#	del mdb.models['Model-1'].parts['Part-1'].sectionAssignments[0]
	# Build sections and assign solid section
	# Define materials and element sets
	for i in range(1,N+1):
		mdl.Material('Material'+str(i)).Elastic(((E[i-1], possoin),))
		mdl.materials['Material'+str(i)].Density(((D[i-1],),))
		mdl.materials['Material'+str(i)].Damping(alpha=alpha, beta=beta)
		mdl.HomogeneousShellSection(name='Sec'+str(i),material='Material'+str(i),thickness=th)
		part.SectionAssignment(part.Set('Set'+str(i),part.elements[i-1:i]),'Sec'+str(i))
	# Define Steps
	mdl.steps['Step-1'].setValues(frequencyRange=((w, w, 2, 1.0), ))
	
def Run_FEA(mdb,Iter):
	# Job Submit
	mdb.Job(name='Design_Job'+str(Iter), model='Model-1')
	a=os.system('del Design_Job'+str(Iter)+'*')
	mdb.jobs['Design_Job'+str(Iter)].writeInput(consistencyChecking=OFF)
	# Modify INP for Gross Matrix
	#f=open('Design_Job'+str(Iter)+'.inp','r+')
	#E_line='Start'
	#while E_line<>'':
	#	E_line = f.readline()
	#	E_line=E_line.strip()
	#	if E_line=='*End Step':
	#		f.seek(-11,2)
	#		f.write("*END STEP\n*STEP, NAME=MATRIX\n*MATRIX GENERATE, STIFFNESS, MASS, VISCOUS DAMPING, LOAD\n\
	#				*Cload, real\nLayer-1.f, 3, -100000.\n*Cload, real\nLayer-1.d1, 3, -0.001\n*Cload, real\nLayer-1.d2, 3, -0.001\n\
	#				*Boundary\nBase-1.b, ENCASTRE\n*MATRIX OUTPUT, STIFFNESS, MASS, VISCOUS DAMPING, LOAD\n*END STEP")
	#		E_line=''
	#f.close()
	# Submit and wait for finishing
	os.system(u'abaqus job='+'Design_Job'+str(Iter))
	FILE='Design_Job'+str(Iter)+'.lck'
	while not os.path.exists(FILE):
		time.sleep(1)		
	while os.path.exists(FILE): 
		time.sleep(1)	 
	return 'Design_Job'+str(Iter)

def Obj_Cal_DISP_READ(U,w,Name):	
	K=np.zeros([441*6,441*6]);M=np.zeros([441*6,441*6]);C=np.zeros([441*6,441*6]);L=np.zeros([441*6,1],dtype=complex)
	Obj=0
	#Reading Gross Matrix---------------------------------------#
	#Reading STIFFNESS MATRIX
	f=open(Name+'_STIF2.mtx','r')
	data_line = f.readline()
	while data_line<>'':
		data_line=data_line.strip().split(',')
		m=(int(data_line[0])-1)*6+int(data_line[1])-1
		n=(int(data_line[2])-1)*6+int(data_line[3])-1
		K[m,n]=float(data_line[4])
		K[n,m]=float(data_line[4])
		data_line = f.readline()
	f.close()
	#Reading MASS MATRIX
	f=open(Name+'_MASS2.mtx','r')
	data_line = f.readline()
	while data_line<>'':
		data_line=data_line.strip().split(',')
		m=(int(data_line[0])-1)*6+int(data_line[1])-1
		n=(int(data_line[2])-1)*6+int(data_line[3])-1
		M[m,n]=float(data_line[4])
		M[n,m]=float(data_line[4])
		data_line = f.readline()
	f.close()
	#Reading DAMPING MATRIX
	f=open(Name+'_DMPV2.mtx','r')
	data_line = f.readline()
	while data_line<>'':
		data_line=data_line.strip().split(',')
		m=(int(data_line[0])-1)*6+int(data_line[1])-1
		n=(int(data_line[2])-1)*6+int(data_line[3])-1
		C[m,n]=float(data_line[4])
		C[n,m]=float(data_line[4])
		data_line = f.readline()
	f.close()
	#--------------------------------------------------------------#
	W=K+1j*w*C-w**2*M
	#Reading LOADING MATRIX
	f=open(Name+'_LOAD2.mtx','r')
	#=======================#
	data_line = f.readline()
	data_line = f.readline()
	#=======================#
	data_line = f.readline()
	while data_line<>'':
		data_line=data_line.strip().split(',')
		nm=int(data_line[0])-1
		nn=int(data_line[1])-1
		nU=U[nm].data[nn]+complex(0,U[nm].conjugateData[nn])
		L[nm*6+nn]=-np.conj(nU)/2.0/np.abs(nU)
		Obj+=np.abs(nU)
		data_line = f.readline()
	f.close()
	mu=np.linalg.solve(W,L)
	return mu.T,Obj

def Node2DOF(Mnodes):
	l1=143;l2=429;l3=572;rem=[]
	for mm in Mnodes:
		if mm <= l1:
			rem.append((data_line[0]-1)*6)
		elif l1< data_line[0] <= l2:
			rem.append(l1*6+((data_line[0])-l1-1)*3)
		elif l2< data_line[0] <= l3:
			rem.append(l1*6+(l2-l1)*3+(data_line[0]-l2-1)*6)
	return np.array(rem,dtype=int)
		
def Read_Gross_Matrix():	
	l1=143;l2=429;l3=572
	GROSS_DOF=6*l1+3*(l2-l1)+6*(l3-l2)
	Name='Job-1'
	K=np.zeros([GROSS_DOF,GROSS_DOF],dtype='f8',order='f');M=np.zeros([GROSS_DOF,GROSS_DOF],dtype='f8',order='f');C=np.zeros([GROSS_DOF,GROSS_DOF],dtype='f8',order='f')
	#Reading Gross Matrix---------------------------------------#
	#Reading STIFFNESS MATRIX
	f=open(Name+'_STIF2.mtx','r')
	data_line = f.readline()
	while data_line<>'':
		data_line=data_line.strip().split(',')
		data_line[0]=int(data_line[0])
		data_line[1]=int(data_line[1])
		data_line[2]=int(data_line[2])
		data_line[3]=int(data_line[3])
		data_line[4]=float(data_line[4])		
		if data_line[0] <= l1:
			m=((data_line[0])-1)*6+data_line[1]-1
		elif l1< data_line[0] <= l2:
			m=l1*6+((data_line[0])-l1-1)*3+data_line[1]-1
		elif l2< data_line[0] <= l3:
			m=l1*6+(l2-l1)*3+((data_line[0])-l2-1)*6+data_line[1]-1
		if data_line[2] <= l1:
			n=((data_line[2])-1)*6+data_line[3]-1
		elif l1< data_line[2] <= l2:
			n=l1*6+((data_line[2])-l1-1)*3+data_line[3]-1
		elif l2< data_line[2] <= l3:
			n=l1*6+(l2-l1)*3+((data_line[2])-l2-1)*6+data_line[3]-1
		K[m,n]=data_line[4]
		K[n,m]=data_line[4]
		data_line = f.readline()
	f.close()
	#Reading MASS MATRIX
	f=open(Name+'_MASS2.mtx','r')
	data_line = f.readline()
	while data_line<>'':
		data_line=data_line.strip().split(',')
		data_line[0]=int(data_line[0])
		data_line[1]=int(data_line[1])
		data_line[2]=int(data_line[2])
		data_line[3]=int(data_line[3])
		data_line[4]=float(data_line[4])		
		if data_line[0] <= l1:
			m=((data_line[0])-1)*6+data_line[1]-1
		elif l1< data_line[0] <= l2:
			m=l1*6+((data_line[0])-l1-1)*3+data_line[1]-1
		elif l2< data_line[0] <= l3:
			m=l1*6+(l2-l1)*3+((data_line[0])-l2-1)*6+data_line[1]-1
		if data_line[2] <= l1:
			n=((data_line[2])-1)*6+data_line[3]-1
		elif l1< data_line[2] <= l2:
			n=l1*6+((data_line[2])-l1-1)*3+data_line[3]-1
		elif l2< data_line[2] <= l3:
			n=l1*6+(l2-l1)*3+((data_line[2])-l2-1)*6+data_line[3]-1
		M[m,n]=data_line[4]
		M[n,m]=data_line[4]
		data_line = f.readline()
	f.close()
	#Reading DAMPING MATRIX
	f=open(Name+'_DMPV2.mtx','r')
	data_line = f.readline()
	while data_line<>'':
		data_line=data_line.strip().split(',')
		data_line[0]=int(data_line[0])
		data_line[1]=int(data_line[1])
		data_line[2]=int(data_line[2])
		data_line[3]=int(data_line[3])
		data_line[4]=float(data_line[4])		
		if data_line[0] <= l1:
			m=((data_line[0])-1)*6+data_line[1]-1
		elif l1< data_line[0] <= l2:
			m=l1*6+((data_line[0])-l1-1)*3+data_line[1]-1
		elif l2< data_line[0] <= l3:
			m=l1*6+(l2-l1)*3+((data_line[0])-l2-1)*6+data_line[1]-1
		if data_line[2] <= l1:
			n=((data_line[2])-1)*6+data_line[3]-1
		elif l1< data_line[2] <= l2:
			n=l1*6+((data_line[2])-l1-1)*3+data_line[3]-1
		elif l2< data_line[2] <= l3:
			n=l1*6+(l2-l1)*3+((data_line[2])-l2-1)*6+data_line[3]-1
		C[m,n]=data_line[4]
		C[n,m]=data_line[4]
		data_line = f.readline()
	f.close()
	#----SETTING BOUNDARY CONDITION----------------------------#
	N=[];N0=range(GROSS_DOF);
	for i in N0:
		if (K[i,i]==1.0e36) or (M[i,i]==0.0):
			N.append(i)
	dN=list(set(N0)-set(N))
	Ld=len(dN)
	K=K[np.ix_(dN,dN)];M=M[np.ix_(dN,dN)];C=C[np.ix_(dN,dN)]
	K0=K.copy();M0=M.copy()
	#load dynamic library
	flib = load_library("test","./")
	fun_1=flib.eigencal
	# shape of 2d array
	# create an empty 2d array
	Z= np.empty(shape=(Ld,Ld),dtype='f8',order='f')
	W= np.empty(shape=(Ld),dtype='f8',order='f')
	info=np.empty(shape=(1),dtype=c_int,order='f')
	# call function
	fun_1.argtypes = [ndpointer(dtype='f8',ndim=2),ndpointer(dtype='f8',ndim=2),ndpointer(dtype='f8',ndim=1),ndpointer(dtype='f8',ndim=2),c_int,c_int,ndpointer(dtype=c_int,ndim=1)]
	flib.eigencal(K0,M0,W,Z,Ld,10,info)
	print "*"*80
	print info
	print np.sqrt(W[0:10])/2.0/np.pi
	for i in range(10):
		print np.dot(np.dot(Z[:,i],C),Z[:,i])/np.dot(np.dot(Z[:,i],K),Z[:,i])
		print np.dot(np.dot(Z[:,i],M),Z[:,i])
	
def Obj_Cal_DC(Elmts,Nds,STIF,MASS,w,alpha,beta,Odb_name,Xe,penal,E1,E0,D1,D0):
	w=w*2*np.pi # Hz to Rad/s
	opdb=openOdb(Odb_name+'.odb')
	U=opdb.steps['Step-1'].frames[-1].fieldOutputs['U'].values
	UR=opdb.steps['Step-1'].frames[-1].fieldOutputs['UR'].values
	U_Obj_K=np.zeros(len(Elmts),dtype=complex);U_Obj_K_C=np.zeros(len(Elmts),dtype=complex)
	U_Obj_M=np.zeros(len(Elmts),dtype=complex);U_Obj_M_C=np.zeros(len(Elmts),dtype=complex)
	Obj_Layer=np.zeros(len(Elmts),dtype=complex)
	Obj_Base=np.zeros(len(Elmts),dtype=complex)
	Obj=np.zeros(len(Elmts),dtype=complex)
	for i in range(len(Elmts)):
		Ue=[]
		for ec in Elmts[i].connectivity:
			j=Nds[ec].label-1
			Ue.extend(U[j].data+1j*U[j].conjugateData)
			Ue.extend(UR[j].data+1j*UR[j].conjugateData)
		Ue=np.array([Ue],dtype=complex)
		S_K=(1+1j*w*beta)*STIF[i+1]
		U_Obj_K[i]=(np.dot(np.dot(Ue,S_K),Ue.T))[0][0]
		U_Obj_K_C[i]=(np.dot(np.dot(Ue,S_K),np.conj(Ue.T)))[0][0]
		S_M=(1j*w*alpha-w**2)*MASS[i+1]
		U_Obj_M[i]=(np.dot(np.dot(Ue,S_M),Ue.T))[0][0]
		U_Obj_M_C[i]=(np.dot(np.dot(Ue,S_M),np.conj(Ue.T)))[0][0]
	# Calculate Objective
	Obj_K=U_Obj_K*(Xe**penal*E1+(1-Xe**penal)*E0)
	Obj_M=U_Obj_M*(Xe**penal*D1+(1-Xe**penal)*D0)
	Obj_Layer=np.sum(Obj_K+Obj_M)
	# Process sensitivities		
	Obj_K_C=U_Obj_K_C*(Xe**penal*E1+(1-Xe**penal)*E0)
	Obj_M_C=U_Obj_M_C*(Xe**penal*D1+(1-Xe**penal)*D0)
	Ae=(np.sum(Obj_K_C+Obj_M_C))
	Ae=-np.real(Ae*penal*Xe**(penal-1)*(U_Obj_M*(D1-D0)+U_Obj_K*(E1-E0)))	
	opdb.close()
	Obj=np.abs(Obj_Layer)
	Ae=Ae/Obj
	return Obj,Ae
	
def Obj_Cal_PF(Elmts,Nds,STIF,MASS,w,alpha,beta,Odb_name,Xe,penal,E1,E0,D1,D0):
	w=w*2*np.pi # Hz to Rad/s
	opdb=openOdb(Odb_name+'.odb')
	V=opdb.steps['Step-1'].frames[-1].fieldOutputs['V'].values
	VR=opdb.steps['Step-1'].frames[-1].fieldOutputs['VR'].values
	CF=opdb.steps['Step-1'].frames[-1].fieldOutputs['CF'].values
	CM=opdb.steps['Step-1'].frames[-1].fieldOutputs['CM'].values
	Sen_K=np.zeros(len(Elmts),dtype=complex)
	Sen_M=np.zeros(len(Elmts),dtype=complex)
	Obj=np.zeros(len(Elmts))
	for i in range(len(Elmts)):
		Ve=[];Fe=[]
		for ec in Elmts[i].connectivity:
			j=Nds[ec].label-1		
			for k in range(len(V[j].data)):
				Ve.append(V[j].data[k]+complex(0,V[j].conjugateData[k]))
			for k in range(len(VR[j].data)):
				Ve.append(VR[j].data[k]+complex(0,VR[j].conjugateData[k]))
			for k in range(len(CF[j].data)):
				Fe.append(CF[j].data[k]+complex(0,CF[j].conjugateData[k]))
			for k in range(len(CM[j].data)):
				Fe.append(CM[j].data[k]+complex(0,CM[j].conjugateData[k]))		
		Ve=np.array([Ve],dtype=complex)
		Fe=np.array([Fe],dtype=complex)
		S_K=(1/(1j*w)+beta)*STIF[i+1]
		Sen_K[i]=(np.dot(np.dot(np.conj(Ve),S_K),Ve.T))[0][0]
		S_M=(1j*w+alpha)*MASS[i+1]
		Sen_M[i]=(np.dot(np.dot(np.conj(Ve),S_M),Ve.T))[0][0]
		# Calculate Obj
		Obj[i]=0.5*np.real(np.dot(np.conj(Fe),Ve.T))
	opdb.close()
	# Calculate Sensitivity
	Ae=-0.5*np.real(penal*Xe**(penal-1)*(Sen_M*(D1-D0)+Sen_K*(E1-E0)))
	return np.sum(Obj),Ae

def Obj_Cal_DISP(Elmts,Nds,STIF,MASS,w,alpha,beta,Odb_name,Xe,penal,E1,E0,D1,D0):
	w=w*2*np.pi # Hz to Rad/s
	opdb=openOdb(Odb_name+'.odb')
	U=opdb.steps['Step-1'].frames[-1].fieldOutputs['U'].values
	UR=opdb.steps['Step-1'].frames[-1].fieldOutputs['UR'].values
	#-----------------------------------------------------------#
	Mu,Obj=Obj_Cal_DISP_READ(U,w,Odb_name)
	#-----------------------------------------------------------#
	U_Obj_K=np.zeros(len(Elmts),dtype=complex)
	U_Obj_M=np.zeros(len(Elmts),dtype=complex)
	for i in range(len(Elmts)):
		Ue=[];Mue=[]
		for ec in Elmts[i].connectivity:
			j=Nds[ec].label-1
			for k in range(len(U[j].data)):
				Ue.append(U[j].data[k]+complex(0,U[j].conjugateData[k]))
			for k in range(len(UR[j].data)):
				Ue.append(UR[j].data[k]+complex(0,UR[j].conjugateData[k]))
			Mue.extend(Mu[0,ec*6:ec*6+6])
		Ue=np.array([Ue],dtype=complex)
		Mue=np.array([Mue],dtype=complex)
		S_K=(1+1j*w*beta)*STIF[i+1]		
		S_M=(1j*w*alpha-w**2)*MASS[i+1]
		U_Obj_K[i]=(np.dot(np.dot(Mue,S_K),Ue.T))[0][0]
		U_Obj_M[i]=(np.dot(np.dot(Mue,S_M),Ue.T))[0][0]
	# Process sensitivities		
	Ae=2*np.real(penal*Xe**(penal-1)*(U_Obj_M*(D1-D0)+U_Obj_K*(E1-E0)))	
	return Obj,Ae

def Obj_Cal_MDR(Elmts,Nds,STIF,MASS,w,alpha,beta,Odb_name,Xe,penal,E1,E0,D1,D0):
	w=w*2*np.pi # Hz to Rad/s
	opdb=openOdb(Odb_name+'.odb')
	U=opdb.steps['Step-1'].frames[-1].fieldOutputs['U'].values
	UR=opdb.steps['Step-1'].frames[-1].fieldOutputs['UR'].values
	#-----------------------------------------------------------#
	Mu,Obj=Read_Gross_Matrix(U,w,Odb_name)
	#-----------------------------------------------------------#
	U_Obj_K=np.zeros(len(Elmts),dtype=complex)
	U_Obj_M=np.zeros(len(Elmts),dtype=complex)
	for i in range(len(Elmts)):
		Ue=[];Mue=[]
		for ec in Elmts[i].connectivity:
			j=Nds[ec].label-1
			for k in range(len(U[j].data)):
				Ue.append(U[j].data[k]+complex(0,U[j].conjugateData[k]))
			for k in range(len(UR[j].data)):
				Ue.append(UR[j].data[k]+complex(0,UR[j].conjugateData[k]))
			Mue.extend(Mu[0,ec*6:ec*6+6])
		Ue=np.array([Ue],dtype=complex)
		Mue=np.array([Mue],dtype=complex)
		S_K=(1+1j*w*beta)*STIF[i+1]		
		S_M=(1j*w*alpha-w**2)*MASS[i+1]
		U_Obj_K[i]=(np.dot(np.dot(Mue,S_K),Ue.T))[0][0]
		U_Obj_M[i]=(np.dot(np.dot(Mue,S_M),Ue.T))[0][0]
	# Process sensitivities		
	Ae=2*np.real(penal*Xe**(penal-1)*(U_Obj_M*(D1-D0)+U_Obj_K*(E1-E0)))	
	return Obj,Ae
	
def Pre_Matrix():
	#--------------Preparation for ESM.mtx-----------------
	# Run original script by abaqus
	#execfile('C:/Temp/FE_Pre.py')
	# Modify INP file and Run for Matrices
	Modify_INP('Job-1')
	os.system(u'abaqus job=Job-1')

def Result_Plot_BESO():
	# Display final design
	vp3 = session.Viewport('Final design', origin=(30,30),width=150, height=100)
	p = mdb.models['Model-1'].parts['Part-1']
	vp3.setValues(displayedObject=p)
	vp3.partDisplay.setValues(mesh=ON)
	vp3.partDisplay.displayGroup.remove(leaf=dgm.LeafFromSets(sets=(p.sets['vs'],)))

def Result_Plot_SIMP(mdb,Xe,Instance_Name):

	itern = len(mdb.customData.History['obj'])	
	# Plot objective function history
	vp1 = session.Viewport('Objective history',origin=(20,20),width=150,height=100)
	try:
		xyPlot1 = session.XYPlot('Objective function')
	except:
		del session.xyPlots['Objective function']
		xyPlot1 = session.XYPlot('Objective function')	
	chart1 = xyPlot1.charts.values()[0]
	objDat = [(k,mdb.customData.History['obj'][k]) for k in range(itern)]
	xydo = session.XYData('Objective function',objDat)
	chart1.setValues(curvesToPlot=[session.Curve(xydo)])
	chart1.axes1[0].axisData.setValues(title='Iteration')
	chart1.axes2[0].axisData.setValues(title='Objective function')
	vp1.setValues(displayedObject=xyPlot1)	
	# Display final design
	opdb = session.openOdb(name= mdb.jobs.keys()[-1]+'.odb',readOnly=FALSE)
	frame=opdb.steps['Step-1'].frames[-1]
	MaterialDensity= frame.FieldOutput(name='MaterialDensity', description='Element Materials Density',type=SCALAR)
	MyInstance=opdb.rootAssembly.instances[Instance_Name]
	elementLabels=range(1,len(MyInstance.elements)+1)
	elementValues=[[i]  for i in Xe]
	MaterialDensity.addData(position=WHOLE_ELEMENT, instance=MyInstance, labels=elementLabels, data=elementValues)
	opdb.save()
	opdb.close()
	
def Cal_Identity(Elmts,w):
	w=w*2*np.pi # Hz to Rad/s
	opdb = session.openOdb(name= mdb.jobs.keys()[-1]+'.odb',readOnly=FALSE)
	Ix=np.zeros(len(Elmts));Iy=np.zeros(len(Elmts));Iz=np.zeros(len(Elmts))
	U=opdb.steps['Step-1'].frames[-1].fieldOutputs['U'].values
	UR=opdb.steps['Step-1'].frames[-1].fieldOutputs['UR'].values
	SF=opdb.steps['Step-1'].frames[-1].fieldOutputs['SF'].values
	SM=opdb.steps['Step-1'].frames[-1].fieldOutputs['SM'].values
	for i in range(len(Elmts)):
		Ue=[];Ixe=[];Iye=[];SFe=[];SMe=[];T=np.zeros([6,6],dtype=complex)	
		for n in range(4):			
			# Ue is conjugateData
			j=Elmts[i].connectivity[n]
			for k in range(len(U[j].data)):
				Ue.append(U[j].data[k]-complex(0,U[j].conjugateData[k]))					
			for k in range(len(UR[j].data)):
				Ue.append(UR[j].data[k]-complex(0,UR[j].conjugateData[k]))		
			if len(SF)==4*len(Elmts):
				j=n+i*4
			else:
				j=i
			for k in range(len(SF[j].data)):
				SFe.append(SF[j].data[k]+complex(0,SF[j].conjugateData[k]))					
			for k in range(len(SM[j].data)):
				SMe.append(SM[j].data[k]+complex(0,SM[j].conjugateData[k]))		
			Ixe.extend([SFe[0],SFe[3],SFe[4],-SMe[2],SMe[1],SFe[2]])
			Iye.extend([SFe[3],SFe[1],SFe[5],-SMe[0],SMe[2],SFe[2]])
		#Coordinate Transfer Matrix T
		T0=np.array(SM[j].localCoordSystem).T
		Tinv=np.linalg.inv(T0)
		T[0:3,0:3]=T0;T[3:6,3:6]=T0
		Ue=np.array([Ue],dtype=complex)
		Ue=Ue.reshape((4,6)).sum(axis=0)/4
		Ue=np.dot(Ue,T) #Coordinate Transfer
		Ixe=np.array([Ixe],dtype=complex)
		Ixe=Ixe.reshape((4,6)).sum(axis=0)/4
		Iye=np.array([Iye],dtype=complex)		
		Iye=Iye.reshape((4,6)).sum(axis=0)/4
		Ide=np.array([-0.5*w*np.imag(np.dot(Ixe,Ue.T)),-0.5*w*np.imag(np.dot(Iye,Ue.T)),0])
		Ix[i],Iy[i],Iz[i]=np.dot(Ide,Tinv)
	frame=opdb.steps['Step-1'].frames[-1]
	while True:
		try:
			StrIdentity= frame.FieldOutput(name='StrIdentity', description='Structural Identity',type=VECTOR, validInvariants=(MAGNITUDE,))
			break
		except:
			StrIdentity= frame.FieldOutput(name='StrIdentity'+str(time.gmtime[5]), description='Structural Identity',type=VECTOR, validInvariants=(MAGNITUDE,))
	MyInstance=opdb.rootAssembly.instances['PART-1-1']
	elementLabels=range(1,len(Elmts)+1)	
	elementValues=[[Ix[i],Iy[i],Iz[i]] for i in range(len(Ix))]
	StrIdentity.addData(position=WHOLE_ELEMENT, instance=MyInstance, labels=elementLabels, data=elementValues)
	opdb.save()
	opdb.close()

def FinalRun(mdb, Xe, E0, E1, D0, D1, penal, Ele_NUM, possoin):
	# Final Run
	# Updating E, D, Obj, Volume 
	E=Xe**penal*E1+(1-Xe**penal)*E0
	D=Xe**penal*D1+(1-Xe**penal)*D0
	# Assign solid and void elements to each section
	for i in range(Ele_NUM):
		mdb.models['Model-1'].materials['Material'+str(i+1)].Elastic(((E[i], possoin),))
		mdb.models['Model-1'].materials['Material'+str(i+1)].Density(((D[i],),))
	mdb.Job('Final_Design','Model-1').submit()
	mdb.jobs['Final_Design'].waitForCompletion()
	
def Topot_Main(Limit_iter):
	global mdb, elmts, nds, Ele_NUM 
	global Part_Name, Instance_Name, ESM_Name
	global vf, rmin, penal, change, iter_num, Xei, vh, oh, Fm 
	global E0, E1, D0, D1, w, alpha, beta, th, possoin
	global Xe, upp, low
	# Start Time
	to=time.time()
	# Close Output Databases
	for i in range(len(session.odbs.keys())):
		session.odbs[session.odbs.keys()[0]].close()
	print 'VolFrac:{0} Rmin:{1} Penal:{2}  X_initial:{3}'.format(vf,rmin,penal,Xei)  
	print 'E_Strong:{0:.3e} E_Weak:{1:.3e} D_Strong:{2} D_Weak:{3}'.format(E1, E0, D1, D0)
	print 'alpha:{0} beta:{1} wp:{2}Hz'.format(alpha,beta,w)
	#-------------Reading Matrix-------------
	STIF={};MASS={};NODENUM={}
	Read_ESM(ESM_Name,STIF,MASS,NODENUM,24)	
	print 'READ {0} ELEMENT MATRICES IN TOTAL'.format(len(STIF))
	#--------------FE Settings--------------
	# Updating E, D
	E=Xe**penal*E1+(1-Xe**penal)*E0
	D=Xe**penal*D1+(1-Xe**penal)*D0		
	FE_Mdb(mdb,E,D,Ele_NUM,w,alpha,beta,th,possoin)
	print "Optimization Start..."	
	# MMA INITIAL PARAMETERS
	m=1;n=Ele_NUM;c0 = 1000;d = 0;a0 = 1;a = 0;
	xold1=np.zeros(n,dtype='f8');xold2=np.zeros(n,dtype='f8')
	xmin=0.001*np.ones(n,dtype='f8');xmax=np.ones(n,dtype='f8');Xmma=np.ones(n,dtype='f8')
	# Load dynamic library
	flib = load_library("mmsub","./");fmmsub=flib.mmsub
	fmmsub.argtypes = [ndpointer(dtype='f8'),ndpointer(dtype='f8'),ndpointer(dtype='f8'),c_int,c_int,c_int,\
	ndpointer(dtype='f8'),ndpointer(dtype='f8'),ndpointer(dtype='f8'),ndpointer(dtype='f8'),ndpointer(dtype='f8'),\
	c_double,ndpointer(dtype='f8'),c_double,ndpointer(dtype='f8'),c_double,c_double,c_double,c_double]	
	# Optimization iteration
	while change > 0.001 and iter_num < Limit_iter:
		# Run FEA
		odb_name=Run_FEA(mdb,iter_num)
		vh.append(np.sum(Xe)/Ele_NUM)
		# Calculate Objective and sensitivities
		#Obj,Ae=Obj_Cal_DISP(elmts,nds,STIF,MASS,w,alpha,beta,odb_name,Xe,penal,E1,E0,D1,D0)
		Obj,Ae=Obj_Cal_DC(elmts,nds,STIF,MASS,w,alpha,beta,odb_name,Xe,penal,E1,E0,D1,D0)
		#Obj,Ae=Obj_Cal_PF(elmts,nds,STIF,MASS,w,alpha,beta,odb_name,Xe,penal,E1,E0,D1,D0)
		oh.append(Obj)
		Ae=fltAe(Ae,Fm,Xe)
		# SIMP optimization
		#Xe=OC(vf,Xe,Ae,Ele_NUM)
		# MMA Algorithm
		f0val=Obj;df0dx=np.array(Ae,dtype='f8');fval=np.sum(Xe)-vf*Ele_NUM;dfdx=np.ones(n,dtype='f8')
		fmmsub(Xmma,low,upp,m,n,iter_num,Xe,xmin,xmax,xold1,xold2,f0val,df0dx,fval,dfdx,a0,a,c0,d)
		xold2 = xold1;xold1 = Xe;Xe=Xmma.copy();
		# Updating E, D, Obj, Volume 
		E=Xe**penal*E1+(1-Xe**penal)*E0
		D=Xe**penal*D1+(1-Xe**penal)*D0
		# Assign solid and void elements to each section
		for i in range(Ele_NUM):
			mdb.models['Model-1'].materials['Material'+str(i+1)].Elastic(((E[i],possoin),))
			mdb.models['Model-1'].materials['Material'+str(i+1)].Density(((D[i],),))		
		change=np.max(np.abs(Xe-xold1))
		if oh[-1] <1e16:
			iter_num += 1
			print 'iter: {3} vol:{0:.2f}, obj:{1}, change:{2:.6f}'.format(vh[-1],oh[-1],change,iter_num)

		else:
			print 'Optimization faild for extreamly large obj.';break 
	
	print "Optimization Finished!"
	print 'Time comsumed:{0}s'.format(time.time()-to)
	# Save results
	mdb.customData.History = {'vol':vh,'obj':oh}
	# Postprocess
	y=sorted(Xe,reverse=True)
	vt=y[int(np.floor((vf-0.001)*Ele_NUM/(1-0.001)))]
	Exa=np.where(Xe>=vt)
	Shr=np.where(Xe<vt)
	Xe[Exa]=1.0;Xe[Shr]=0.001
	# Display
	Result_Plot_SIMP(mdb,Xe,Instance_Name)
	#FinalRun(mdb, Xe, E0, E1, D0, D1, penal, Ele_NUM, possoin)
	# Cal Structural Identity
	#Cal_Identity(elmts,w)
	pass
	 
print 'Load Pramameters'
Model_Name='Plate1600.cae'
Part_Name='Part-1'
Instance_Name='PART-1-1'
ESM_Name='ESM1600.mtx'
# Define Optimization Parameters
vf, rmin, penal, change, iter_num, Xei = [0.5, 0.02, 3, 1, 0, 0.5]
E0, E1, D0, D1 = [1e10, 1e11, 780, 7800 ]
th, possoin, alpha, beta=[0.003,0.3, 1E-4, 1E-5] #thickness, possoin ratio, damping
w=693 # Excitation Frequency (Hz cycles/s)
vh=[];oh=[]#History Restore
# Model Setting
mdb=openMdb(Model_Name)
elmts=mdb.models['Model-1'].parts[Part_Name].elements
nds =mdb.models['Model-1'].parts[Part_Name].nodes
Ele_NUM=len(elmts)
# MMA parameters
Xe=Xei*np.ones(Ele_NUM,dtype='f8')
low=np.zeros(Ele_NUM,dtype='f8')
upp=np.zeros(Ele_NUM,dtype='f8')
Fm=preFlt(rmin,elmts,nds)
# Run Topology optimization
Topot_Main(60)
