%%%% AN 88 LINE TOPOLOGY OPTIMIZATION CODE Nov, 2010 %%%%
%% defult
clear;
volfrac=0.5;penal=3;rmin=0.02;
%% MATERIAL PROPERTIES
Es1=68.9e9;G2=0.896e6*(1+0.5i);Es3=68.9e9;p1=2740;p2=999;p3=2740;
%% PREPARE FINITE ELEMENT ANALYSIS
nelx=20;nely=20;nelz=1;
h1=0.762*10^-3;h2=0.254*10^-3;h3=(0.762)*10^-3;
Lx=0.3048;Ly=0.3480;Lz=h1+h2+h3;
[~,Ee]=ConstructModel(Lx,nelx,Ly,nely,Lz,nelz);
%% Build Element Matrix
load K&M.mat
K_cal=matlabFunction(K);
M_cal=matlabFunction(M);
%% INITIALIZE ITERATION
x = volfrac*ones(nelx*nely,1);
r=rmin/(Lx/nelx);
%% PREPARE FILTER
iH = ones(nelx*nely*(2*(ceil(r)-1)+1)^2,1);
jH = ones(size(iH));
sH = zeros(size(iH));
k = 0;
sx=zeros(nelx,nely);
for i1 = 1:nelx
    for j1 = 1:nely
        le=(i1-1)*nely+j1; sx(i1,j1)=le;
        for i2 = max(i1-(ceil(r)-1),1):min(i1+(ceil(r)-1),nelx)
            for j2 = max(j1-(ceil(r)-1),1):min(j1+(ceil(r)-1),nely)
                lf=(i2-1)*nely+j2;
                k = k+1;
                iH(k) = le;
                jH(k) = lf;
                sH(k) = max(0,r-sqrt((i1-i2)^2+(j1-j2)^2));
            end
        end
    end
end
sH=sH*Lx/nelx;
H = sparse(iH,jH,sH);
Hs = sum(H,2);
%% MMA INITIAL PARAMETERS
m=1;if nelz==0,n=nelx*nely;else n=nelx*nely*nelz;end
xold1=zeros(n,1);xold2=zeros(n,1);
low=zeros(n,1);upp=zeros(n,1);
xmin=0.000*ones(n,1);xmax=ones(n,1);
c0 = 1e4;d = 0;a0 = 1;a = 0;
iter=0;change = 1;Obj=0;
Oh=zeros(40,3);
%% START ITERATION
while abs(change) > 0.005 && iter<48
    xPhys=x;ObjPhys=Obj;    
    %% FE-ANALYSIS OBJECTIVE FUNCTION AND SENSITIVITY ANALYSIS
    [K,M,Z_all,W_all,dN,Lf_all,edofMat]=FESolve(Es1,G2,Es3,p1,p2,p3,Ee,x,K_cal,M_cal,penal,nelx,nely,nelz);
    Oh(iter+1,:)=[Lf_all(1),Lf_all(2),Lf_all(3)];
    [Obj,dc]=Obj_MDR(Es1,G2,Es3,p1,p2,p3,W_all,Z_all,K,M,Lf_all,dN,edofMat,x,K_cal,M_cal,penal);
    %% MMA UPDATE OF DESIGN VARIABLES AND PHYSICAL DENSITIES
    iter = iter+1;
    dc = H*(x.*dc)./Hs./max(1e-3,x);% FILTERING/MODIFICATION OF SENSITIVITIES
    %% MMA
    f0val=Obj;df0dx=dc;fval=sum(x)-(volfrac*nelx*nely);dfdx=ones(1,n);
    [xmma,ymma,zmma,lam,xsi,eta,mu,zet,s,low,upp] = mmasub(m,n,iter,x,xmin,xmax,xold1,xold2,f0val,df0dx,fval,dfdx,low,upp,a0,a,c0,d);
    xold2 = xold1;
    xold1 = x;
    x = xmma;
    %% PRINT RESULTS
    change=sum(x)-(volfrac*nelx*nely);
    fprintf(' It.:%4i Obj.:%7.6f Cov.:%7.3f Vol.:%7.3f ch.: %7.5f\n',iter,Obj,change,sum(x)/(nelx*nely),abs((Obj-ObjPhys)/ObjPhys));
    %% PLOT DENSITIES
    sxx=real(x(sx));
    figure(3)
    colormap(gray); imagesc(1-sxx); caxis([0 1]); axis equal; axis off; drawnow;

end



