from caeModules import *
import math, customKernel
from abaqusConstants import *
from Parameters import *
from odbAccess import openOdb

def preFlt(Rmin,Elmts,Nds):
	Fm={}
	# Calculate element centre coordinates
	elm, c0 = np.zeros(len(Elmts)), np.zeros((len(Elmts),3))
	for i in range(len(elm)):
		elm[i] = Elmts[i].label
		nds = Elmts[i].connectivity
		for nd in nds: c0[i] = np.add(c0[i],np.divide(Nds[nd].coordinates,len(nds)))
	# Weighting factors
	for i in range(len(elm)):
		Fm[elm[i]] = [[],[]]
		for j in range(len(elm)):
			dis = np.sqrt(np.sum(np.power(np.subtract(c0[i],c0[j]),2)))
			if dis<Rmin:
				Fm[elm[i]][0].append(elm[j])
				Fm[elm[i]][1].append(Rmin - dis)
		Fm[elm[i]][1] = np.divide(Fm[elm[i]][1],np.sum(Fm[elm[i]][1]))
	return Fm

def Post_process(Xe,Ele_NUM,vf):
	y=sorted(Xe,reverse=True)
	vt=y[int(np.floor((vf-0.001)*Ele_NUM/(1-0.001)))]
	Exa=np.where(Xe>=vt)
	Shr=np.where(Xe<vt)
	Xe[Exa]=1.0;Xe[Shr]=0.001
	return Xe

def Modify_INP():
	# Modify INP for Gross Matrix
	f=open('Pre_Run.inp','r+')
	E_line='Start'
	while E_line<>'':
		E_line = f.readline()
		E_line=E_line.strip()
		if E_line=='*End Step':
			f.seek(-11,2)
			f.write("\
*END STEP\n\
*STEP, NAME=MATRIX\n\
*MATRIX GENERATE, STIFFNESS, MASS, VISCOUS DAMPING, LOAD\n\
*Boundary\n\
Base-1.b, ENCASTRE\n\
*Dsload, real\n\
Layer-1.up, P, 1e-06\n\
*MATRIX OUTPUT, STIFFNESS, MASS, VISCOUS DAMPING, LOAD\n\
*END STEP")
			E_line=''
	f.close()
	
def InitialXe(mdb,N):
	# Generate Initial Xe
	mdb.Job('Pre_Run','Model-1').submit()
	mdb.jobs['Pre_Run'].waitForCompletion()
	opdb=openOdb('Pre_Run.odb')
	U=opdb.steps['Step-1'].frames[-1].fieldOutputs['U'].values
	elmts=mdb.models['Model-1'].parts[Part_Name[1]].elements
	Xe=np.zeros(N)
	for i in range(len(elmts)):
		Ue=[];
		for j in elmts[i].connectivity:			
			Ue.append(U[j].data[2]+1j*U[j].conjugateData[2])
		Xe[i]=abs(np.sum(np.array(Ue))/4)
	opdb.close()
	Xe=Post_process(Xe,N,vf)
	f=open('Xe_Initial.txt','w')
	pickle.dump(Xe,f)
	f.close()
	return Xe
	
def FE_Mdb(mdb,Mats,N,w,Xe,PN):
	# Parameters Define
	mdl = mdb.models['Model-1'];part = mdl.parts[PN]
	E0=Mats[PN]['E0'];E1=Mats[PN]['E1'];D0=Mats[PN]['D0'];D1=Mats[PN]['D1']
	alpha=Mats[PN]['alpha'];beta=Mats[PN]['beta'];th=Mats[PN]['th'];possion=Mats[PN]['possion']
	E=Xe**K_penal*E1+(1-Xe**K_penal)*E0
	D=Xe**M_penal*D1+(1-Xe**M_penal)*D0	
	for i in range(1,N+1):
		mdl.Material('Material'+PN+str(i)).Elastic(((E[i-1], possion),))
		mdl.materials['Material'+PN+str(i)].Density(((D[i-1],),))
		mdl.materials['Material'+PN+str(i)].Damping(alpha=alpha, beta=beta)
		mdl.HomogeneousShellSection(name='Sec'+PN+str(i),material='Material'+PN+str(i),thickness=th)
		part.SectionAssignment(part.Set('Set'+PN+str(i),part.elements[i-1:i]),'Sec'+PN+str(i))
	mdl.steps['Step-1'].setValues(frequencyRange=((w, w, 2, 1.0), ))
	# Job Submit
	os.system('del Pre_Run*');os.system('del Design_Job*')
	mdb.Job(name='Pre_Run', model='Model-1')
	mdb.jobs['Pre_Run'].writeInput(consistencyChecking=OFF)	

# Model Setting
mdb=openMdb(Model_Name)
elmts=mdb.models['Model-1'].parts[Part_Name[1]].elements
nds =mdb.models['Model-1'].parts[Part_Name[1]].nodes
Ele_NUM=len(elmts)
f=open('log.txt','w')
f.write('Total Elements Number:'+str(Ele_NUM)+'\n')
f.write('Total Elements Number:'+str(len(nds))+'\n')	
Xe=Xei*np.ones(Ele_NUM)
FE_Mdb(mdb,Mats,Ele_NUM,w,Xe,Part_Name[1])
Xe=InitialXe(mdb,Ele_NUM)
# Modify_INP
Modify_INP()
# Filter Preparation
if not os.path.exists('Fm3600.txt'):
	FLAG='Y'
else:
	FLAG='N'
f.write('SENSITIVITY FILTER RADIUS: '+str(rmin)+'\n')
f.write('SHIFT_PARA: '+FLAG)
f.close()
if FLAG<>'N':
	Fm=preFlt(rmin,elmts,nds)
	f=open('Fm'+str(Ele_NUM)+'.txt','w')
	pickle.dump(Fm,f)
	f.close()