# Verison 1.2
# Add Read_ESM function used to read STIFFNESS and MASS matrices from files.
# Stop obtaining Element Elastic Energy from ABAQUS, Calculate Objective by external function 
# Delete Resubmit Function for it's useless. The reason, which is not sure enough, why ABAQUS collapse is MemoryError
# Adopt SIMP model which means OC algorithm and sensitivity filter were introduced in this version
# Add some help functions: Modify INP, Pre_Matrix, Result_Plot_SIMP
#Version 1.3
# SIMP Bi-Mateiral Model Static Optimization
# Add Dynamic Optimization
# Modify Obj_Cal for complex calculation
# Fix a bug for Read_ESM function. It can't read MASS matrix before.
# Fix a bug in preFlt function. distance of two elements should be sqrt of the sum of the squares of coordinates.
# Add Self-define excitation frequency of load
# Modify OC algorithm for Non-Negative in sqrt function
#Version 1.4
# Add Power-flow Objective function and redesign code structure which enable the code more compatible for different Objective function.
# Fix excitation frequency. Should be natural(radian) frequency NOT cycles frequency
# Add MMA Algorithm
# Add Structural Intensity Function
# Fix Structural Intensity Function under Global Coordinate System
#Version 1.5
# Add Restore Mechanism
# Add Post-process 
# Revise the programming structure. Add global parameters setting
# Add Final Run
# Revise restore mechanism by global variables
#Version 1.6
# Add Displacement Objective
# Add Free Damping Layer Model
# Change the FEA run mode by command line
# Add GROSS MATRIX READ function
# Read Node number as well as the Element Matrix
#Version 1.6_Optimization
# Build data connection with ABAQUS by .txt.So a huge step to run code externally which means all newest python packages are available for calculation.
# Using Sparse matrix in GROSS MATRIX READ to overcome memory issue.
# Update Material Parameters by modifying INP directly.
# Introduce Global variables modulus "Parameters"

from scipy.sparse import csr_matrix
from scipy.sparse.linalg import spsolve
from Parameters import *

def Read_ESM(ESM_Name, STIF, MASS, NODENUM, N):
	f=open(ESM_Name,'r')
	E_line='Start'
	E_NUM=0    
	while E_line<>'':
		E_line = f.readline()
		if E_line[3:17]=='ELEMENT NUMBER':
			E_NUM=E_NUM+1
		elif E_line[3:16]=='ELEMENT NODES':
			E_line=f.readline();
			NODENUM[E_NUM]=np.array(E_line[2:-1].strip().split(','),dtype=int)
		Head=E_line[8:len(E_line)-1].strip()
		if Head=='TYPE=STIFFNESS' and E_NUM<>0:
			STIF[E_NUM],MASS[E_NUM]=Read_Matrices(f,N,Head[5:len(Head)],E_NUM)
	f.close()
    
def Read_Matrices(f,N,Type,Enum):
	Line_Matrix=[];K=np.zeros([N,N]);M=np.zeros([N,N])
	data_line = f.readline()
	data_line=data_line.strip().split(',');data_line.pop() 
	while data_line[0]<>'**':
		Line_Matrix.extend(data_line)
		data_line = f.readline()
		if data_line=='':break
		data_line=data_line.strip().split(',')
		if data_line[-1]=='': data_line.pop()  
	Line_Matrix.reverse()
	for i in range(1,1+N):    
		for j in range(1,1+i):
			K[i-1,j-1]=Line_Matrix.pop()
			if i<>j:K[j-1,i-1]=K[i-1,j-1]
	Line_Matrix.pop();Line_Matrix.pop()
	for i in range(1,1+N):    
		for j in range(1,1+i):
			M[i-1,j-1]=Line_Matrix.pop()
			if i<>j:M[j-1,i-1]=M[i-1,j-1]
	return K,M
    
def fltAe(Ae,Fm,Xe):
	raw = Ae.copy()
	for el in Fm.keys():
		Ae[el-1] = 0.0
		for i in range(len(Fm[el][0])): Ae[el-1]+=raw[Fm[el][0][i]-1]*Fm[el][1][i]*Xe[Fm[el][0][i]-1]/Xe[el-1]
	return Ae

def FE_Mdb(FM_Name):
	os.system("del Design_Job0.inp")
	os.system("abaqus cae noGUI=a_1.6(SIMP)_Abaqus.py")
	FILE=FM_Name
	while not os.path.exists(FILE):
		time.sleep(1)	
	time.sleep(1)
	shutil.copyfile('Pre_Run.inp', 'Design_Job0.inp')
	os.system("del Pre_Run*")

def Run_FEA(Xe,Mats,Part_Name,Iter):	
	# Job Submit
	Modify_INP(Xe,Mats,Iter,Part_Name)
	Iter=Iter+1
	os.system(u'abaqus job='+'Design_Job'+str(Iter))
	FILE='Design_Job'+str(Iter)+'.lck'
	while not os.path.exists(FILE):
		time.sleep(1)		
	while os.path.exists(FILE): 
		time.sleep(1)
	os.system(u'abaqus python a_1.6(SIMP)_Output.py'+' -- Design_Job'+str(Iter))
	FILE='Design_Job'+str(Iter)+'_U.txt'
	while not os.path.exists(FILE):
		time.sleep(1)
	FILE='Design_Job'+str(Iter)+'_UR.txt'
	while not os.path.exists(FILE):
		time.sleep(1)
	time.sleep(1)
	return 'Design_Job'+str(Iter),Iter

def Obj_Cal_DISP_READ_SPARSE(U,w,Name):
	K_DATA=[];K_ROW=[];K_COL=[]
	M_DATA=[];M_ROW=[];M_COL=[]
	C_DATA=[];C_ROW=[];C_COL=[]
	L_DATA=[];L_ROW=[];L_COL=[]
	Obj=0
	#Reading STIFFNESS MATRIX
	f=open(Name+'_STIF2.mtx','r')
	data_line = f.readline()
	while data_line<>'':
		data_line=data_line.strip().split(',')
		m=(int(data_line[0])-1)*6+int(data_line[1])-1
		n=(int(data_line[2])-1)*6+int(data_line[3])-1
		K_ROW.extend((m,n))
		K_COL.extend((n,m))
		if m==n:
			data=float(data_line[4])/2
		else:
			data=float(data_line[4])
		K_DATA.extend((data,data))
		data_line = f.readline()
	f.close()
	K=csr_matrix((K_DATA,(K_ROW,K_COL)), shape=(np.max(K_ROW)+1,np.max(K_COL)+1),dtype=complex)
	#Reading MASS MATRIX
	f=open(Name+'_MASS2.mtx','r')
	data_line = f.readline()
	while data_line<>'':
		data_line=data_line.strip().split(',')
		m=(int(data_line[0])-1)*6+int(data_line[1])-1
		n=(int(data_line[2])-1)*6+int(data_line[3])-1
		M_ROW.extend((m,n))
		M_COL.extend((n,m))
		if m==n:
			data=float(data_line[4])/2
		else:
			data=float(data_line[4])
		M_DATA.extend((data,data))
		data_line = f.readline()
	f.close()
	M=csr_matrix((M_DATA,(M_ROW,M_COL)), shape=(np.max(M_ROW)+1,np.max(M_COL)+1),dtype=complex)
	#Reading DAMPING MATRIX
	f=open(Name+'_DMPV2.mtx','r')
	data_line = f.readline()
	while data_line<>'':
		data_line=data_line.strip().split(',')
		m=(int(data_line[0])-1)*6+int(data_line[1])-1
		n=(int(data_line[2])-1)*6+int(data_line[3])-1
		C_ROW.extend((m,n))
		C_COL.extend((n,m))
		if m==n:
			data=float(data_line[4])/2
		else:
			data=float(data_line[4])
		C_DATA.extend((data,data))
		data_line = f.readline()
	f.close()
	C=csr_matrix((C_DATA,(C_ROW,C_COL)), shape=(np.max(C_ROW)+1,np.max(C_COL)+1),dtype=complex)
	#Reading LOADING MATRIX
	f=open(Name+'_LOAD2.mtx','r')
	#=======================#
	data_line = f.readline()
	data_line = f.readline()
	#=======================#
	data_line = f.readline()
	while data_line<>'':
		data_line=data_line.strip().split(',')
		nm=int(data_line[0])-1;	nn=int(data_line[1])-1;	nU=U[nm][nn]
		L_ROW.append(nm*6+nn);L_COL.append(0);L_DATA.append(-np.conj(nU)/2.0/np.abs(nU))		
		Obj+=np.abs(nU)
		data_line = f.readline()
	f.close()
	L=csr_matrix((L_DATA,(L_ROW,L_COL)), shape=(np.max(K_ROW)+1,1),dtype=complex)
	W=K+1j*w*C-w**2*M
	mu=spsolve(W,L)
	return mu.T,Obj	
	
def Obj_Cal_DC(ESM,w,Odb_name,Xe,penal,Mats,PN,Ele_NUM):
	w=w*2*np.pi # Hz to Rad/s
	f=open(Odb_name+'_U.txt','r')
	U=pickle.load(f)
	f.close()
	f=open(Odb_name+'_UR.txt','r')
	UR=pickle.load(f)
	f.close()
	#------------------------------------------------------------------#
	STIF=ESM['ESMLAYER3600.mtx']['STIF'];MASS=ESM['ESMLAYER3600.mtx']['MASS'];NODENUM=ESM['ESMLAYER3600.mtx']['NODENUM']
	E0=Mats[PN]['E0'];E1=Mats[PN]['E1'];D0=Mats[PN]['D0'];D1=Mats[PN]['D1'];alpha=Mats[PN]['alpha'];beta=Mats[PN]['beta']
	U_Obj_K=np.zeros(Ele_NUM,dtype=complex);U_Obj_K_C=np.zeros(Ele_NUM,dtype=complex)
	U_Obj_M=np.zeros(Ele_NUM,dtype=complex);U_Obj_M_C=np.zeros(Ele_NUM,dtype=complex)
	Obj_Layer=np.zeros(Ele_NUM,dtype=complex)
	Obj_Base=np.zeros(Ele_NUM,dtype=complex)
	Obj=np.zeros(Ele_NUM,dtype=complex)
	#------------------------------------------------------------------#
	for i in range(Ele_NUM):
		Ue=[]
		for ec in NODENUM[i+1]:
			Ue.extend(U[ec-1]);Ue.extend(UR[ec-1])
		Ue=np.array([Ue],dtype=complex)
		S_K=(1+1j*w*beta)*STIF[i+1]
		U_Obj_K[i]=(np.dot(np.dot(Ue,S_K),Ue.T))[0][0]
		U_Obj_K_C[i]=(np.dot(np.dot(Ue,S_K),np.conj(Ue.T)))[0][0]
		S_M=(1j*w*alpha-w**2)*MASS[i+1]
		U_Obj_M[i]=(np.dot(np.dot(Ue,S_M),Ue.T))[0][0]
		U_Obj_M_C[i]=(np.dot(np.dot(Ue,S_M),np.conj(Ue.T)))[0][0]
	# Calculate Objective
	Obj_K=U_Obj_K*(Xe**penal*E1+(1-Xe**penal)*E0)
	Obj_M=U_Obj_M*(Xe**penal*D1+(1-Xe**penal)*D0)
	Obj_Layer=np.sum(Obj_K+Obj_M)
	# Process sensitivities		
	Obj_K_C=U_Obj_K_C*(Xe**penal*E1+(1-Xe**penal)*E0)
	Obj_M_C=U_Obj_M_C*(Xe**penal*D1+(1-Xe**penal)*D0)
	Ae=(np.sum(Obj_K_C+Obj_M_C))
	Ae=-np.real(Ae*penal*Xe**(penal-1)*(U_Obj_M*(D1-D0)+U_Obj_K*(E1-E0)))	
	#------------------------------------------------------------------#
	PN=Part_Name[0]
	STIF=ESM['ESMBASE3600.mtx']['STIF'];MASS=ESM['ESMBASE3600.mtx']['MASS'];NODENUM=ESM['ESMBASE3600.mtx']['NODENUM']
	E0=Mats[PN]['E0'];E1=Mats[PN]['E1'];D0=Mats[PN]['D0'];D1=Mats[PN]['D1'];alpha=Mats[PN]['alpha'];beta=Mats[PN]['beta']
	for i in range(Ele_NUM):
		Ue=[]
		for ec in NODENUM[i+1]:
			Ue.extend(U[ec-1]);Ue.extend(UR[ec-1])
		Ue=np.array([Ue],dtype=complex)
		S_K=(1+1j*w*beta)*STIF[i+1]
		U_Obj_K[i]=(np.dot(np.dot(Ue,S_K),Ue.T))[0][0]
		S_M=(1j*w*alpha-w**2)*MASS[i+1]
		U_Obj_M[i]=(np.dot(np.dot(Ue,S_M),Ue.T))[0][0]
	Obj_Base=np.sum(U_Obj_K+U_Obj_M)
	Obj=np.abs(Obj_Layer+Obj_Base)
	Ae=Ae/Obj
	return Obj,Ae
	
def Obj_Cal_PF(Elmts,Nds,STIF,MASS,w,alpha,beta,Odb_name,Xe,penal,E1,E0,D1,D0):
	w=w*2*np.pi # Hz to Rad/s
	opdb=openOdb(Odb_name+'.odb')
	V=opdb.steps['Step-1'].frames[-1].fieldOutputs['V'].values
	VR=opdb.steps['Step-1'].frames[-1].fieldOutputs['VR'].values
	CF=opdb.steps['Step-1'].frames[-1].fieldOutputs['CF'].values
	CM=opdb.steps['Step-1'].frames[-1].fieldOutputs['CM'].values
	Sen_K=np.zeros(len(Elmts),dtype=complex)
	Sen_M=np.zeros(len(Elmts),dtype=complex)
	Obj=np.zeros(len(Elmts))
	for i in range(len(Elmts)):
		Ve=[];Fe=[]
		for ec in Elmts[i].connectivity:
			j=Nds[ec].label-1		
			for k in range(len(V[j].data)):
				Ve.append(V[j].data[k]+complex(0,V[j].conjugateData[k]))
			for k in range(len(VR[j].data)):
				Ve.append(VR[j].data[k]+complex(0,VR[j].conjugateData[k]))
			for k in range(len(CF[j].data)):
				Fe.append(CF[j].data[k]+complex(0,CF[j].conjugateData[k]))
			for k in range(len(CM[j].data)):
				Fe.append(CM[j].data[k]+complex(0,CM[j].conjugateData[k]))		
		Ve=np.array([Ve],dtype=complex)
		Fe=np.array([Fe],dtype=complex)
		S_K=(1/(1j*w)+beta)*STIF[i+1]
		Sen_K[i]=(np.dot(np.dot(np.conj(Ve),S_K),Ve.T))[0][0]
		S_M=(1j*w+alpha)*MASS[i+1]
		Sen_M[i]=(np.dot(np.dot(np.conj(Ve),S_M),Ve.T))[0][0]
		# Calculate Obj
		Obj[i]=0.5*np.real(np.dot(np.conj(Fe),Ve.T))
	opdb.close()
	# Calculate Sensitivity
	Ae=-0.5*np.real(penal*Xe**(penal-1)*(Sen_M*(D1-D0)+Sen_K*(E1-E0)))
	return np.sum(Obj),Ae

def Obj_Cal_DISP(STIF,MASS,NODENUM,w,Odb_name,Xe,Mats,PN):
	global shift_para
	w=w*2*np.pi # Hz to Rad/s
	f=open(Odb_name+'_U.txt','r')
	U=pickle.load(f)
	f.close()
	f=open(Odb_name+'_UR.txt','r')
	UR=pickle.load(f)
	f.close()
	#-----------------------------------------------------------#
	Mu,Obj=Obj_Cal_DISP_READ_SPARSE(U,w,Odb_name)
	#-----------------------------------------------------------#
	E0=Mats[PN]['E0'];E1=Mats[PN]['E1'];D0=Mats[PN]['D0'];D1=Mats[PN]['D1'];alpha=Mats[PN]['alpha'];beta=Mats[PN]['beta']
	U_Obj_K=np.zeros(len(STIF),dtype=complex)
	U_Obj_M=np.zeros(len(STIF),dtype=complex)
	for i in range(len(STIF)):
		Ue=[];Mue=[]
		for ec in NODENUM[i+1]:
			Ue.extend(U[ec-1]);Ue.extend(UR[ec-1])
			Mue.extend(Mu[(ec-shift_para)*6:(ec-shift_para)*6+6])
		Ue=np.array([Ue],dtype=complex)
		Mue=np.array([Mue],dtype=complex)
		U_Obj_K[i]=(np.dot(np.dot(Mue,STIF[i+1]),Ue.T))[0][0]
		U_Obj_M[i]=(np.dot(np.dot(Mue,MASS[i+1]),Ue.T))[0][0]
	# Process sensitivities	
	S_K=K_penal*Xe**(K_penal-1)*U_Obj_K*(E1-E0)
	S_M=-w**2*M_penal*Xe**(M_penal-1)*U_Obj_M*(D1-D0)
	S_C=1j*w*(alpha*K_penal*Xe**(K_penal-1)*U_Obj_M*(D1-D0)+beta*S_K)
	Ae=2*np.real(S_K+S_M+S_C)
	return Obj,Ae

def Modify_INP(Xe,Mats,iter,Part_Name):
	fr=open("Design_Job0.inp",'r')
	fo=open("Design_Job"+str(iter+1)+".inp",'w')
	E_line='Start'
	while E_line<>'':
		E_line = fr.readline()
		if E_line[0:9]=='*Material':
			if E_line[16:29]=="MaterialLayer":
				N=int(E_line[29:-1])
				PN=Part_Name[1];E0=Mats[PN]['E0'];E1=Mats[PN]['E1'];D0=Mats[PN]['D0'];D1=Mats[PN]['D1'];poisson=Mats[PN]['possion']		
				alpha=Mats[PN]['alpha'];beta=Mats[PN]['beta']
				E=max(Xe[N-1]**K_penal*E1+(1-Xe[N-1]**K_penal)*E0,1e-1)
				D=max(Xe[N-1]**M_penal*D1+(1-Xe[N-1]**M_penal)*D0,1e-3)
				D_damping=max(Xe[N-1]**K_penal*D1+(1-Xe[N-1]**K_penal)*D0,1e-3)
				fo.write(E_line)
				E_line = fr.readline();fo.write('*Damping, alpha='+str(round(alpha*D_damping/D,3))+', beta='+str(beta)+'\n')
				E_line = fr.readline();fo.write(E_line)
				E_line = fr.readline();fo.write(" "+str(round(D,3))+",\n")
				E_line = fr.readline();fo.write(E_line)
				E_line = fr.readline();fo.write(" "+str(round(E,1))+", "+str(poisson)+"\n")
			else:
				fo.write(E_line)
		else:
			fo.write(E_line)
	fr.close()
	fo.close()

def Topot_Main(Limit_iter,Xe,upp,low,change,iter_num):
	global ESM, Ele_NUM 
	global Part_Name, Instance_Name, w, to
	global vh, oh, Fm 
	print "Optimization Start...      Time: "+time.strftime("%H:%M:%S")	
	# MMA INITIAL PARAMETERS
	m=1;n=Ele_NUM;c0 = 1000;d = 0;a0 = 1;a = 0;
	xold1=np.zeros(n,dtype='f8');xold2=np.zeros(n,dtype='f8')
	xmin=0.001*np.ones(n,dtype='f8');xmax=np.ones(n,dtype='f8');Xmma=np.ones(n,dtype='f8')
	# Load dynamic library
	flib = load_library("mmsub","./");fmmsub=flib.mmsub
	fmmsub.argtypes = [ndpointer(dtype='f8'),ndpointer(dtype='f8'),ndpointer(dtype='f8'),c_int,c_int,c_int,\
	ndpointer(dtype='f8'),ndpointer(dtype='f8'),ndpointer(dtype='f8'),ndpointer(dtype='f8'),ndpointer(dtype='f8'),\
	c_double,ndpointer(dtype='f8'),c_double,ndpointer(dtype='f8'),c_double,c_double,c_double,c_double]	
	# Optimization iteration
	g=open('OH.txt','a')
	while change > 0.001 and iter_num < Limit_iter:
		# Run FEA
		print 'Run_FEA_Time comsumed:{0}s'.format(time.time()-to)
		odb_name,iter_num=Run_FEA(Xe,Mats,Part_Name,iter_num)
		print 'Run_FEA_Time comsumed:{0}s'.format(time.time()-to)
		vh.append(np.sum(Xe)/Ele_NUM)
		# Calculate Objective and sensitivities
		Obj,Ae=Obj_Cal_DISP(ESM['ESMLAYER3600.mtx']['STIF'],ESM['ESMLAYER3600.mtx']['MASS'],ESM['ESMLAYER3600.mtx']['NODENUM'],w,odb_name,Xe,Mats,Part_Name[1])
		#Obj,Ae=Obj_Cal_DC(ESM,w,odb_name,Xe,penal,Mats,Part_Name[1],Ele_NUM)
		#Obj,Ae=Obj_Cal_PF(elmts,nds,STIF,MASS,w,alpha,beta,odb_name,Xe,penal,E1,E0,D1,D0)
		oh.append(Obj)
		Ae=fltAe(Ae,Fm,Xe)
		# MMA Algorithm
		f0val=Obj;df0dx=np.array(Ae,dtype='f8');fval=np.sum(Xe)-vf*Ele_NUM;dfdx=np.ones(n,dtype='f8')
		fmmsub(Xmma,low,upp,m,n,iter_num,Xe,xmin,xmax,xold1,xold2,f0val,df0dx,fval,dfdx,a0,a,c0,d)
		xold2 = xold1;xold1 = Xe;Xe=np.round(Xmma.copy(),3);
		change=np.max(np.abs(Xe-xold1))
		if oh[-1] <1e16:
			IterResOut= 'iter: {3} vol:{0:.2f}, obj:{1}, change:{2:.6f}'.format(vh[-1],oh[-1],change,iter_num);print IterResOut
			g.write(IterResOut+'\n')
			f=open('Xe_'+str(iter_num)+'.txt','w')
			pickle.dump(Xe,f)
			pickle.dump(low,f)
			pickle.dump(upp,f)
			f.close()
		else:
			print 'Optimization faild for extreamly large obj.';break	
	g.close()
	print "Optimization Finished!   Time: "+time.strftime("%H:%M:%S")
	print 'Time comsumed:{0}s'.format(time.time()-to)
	# PostProcess
	#os.system('abaqus cae noGUI=a_1.6(SIMP)_Result.py -- '+str(Limit_iter))
	pass
	 
print 'Load Pramameters'
shift_para=3722
vh=[];oh=[];ESM={}#History Restore
# Reading Matrix
for EN in ESM_Name:
	ESM[EN]={};ESM[EN]['STIF']={};ESM[EN]['MASS']={};ESM[EN]['NODENUM']={}
	Read_ESM(EN,ESM[EN]['STIF'],ESM[EN]['MASS'],ESM[EN]['NODENUM'],24)	
	print 'READ {0} ELEMENTS MATRICES IN TOTAL: {1}'.format(len(ESM[EN]['STIF']),EN)
Ele_NUM=len(ESM['ESMLAYER3600.mtx']['STIF'])

iter_num=0
if iter_num==0:
	# Initialization Fm and inp File
	os.system('del OH.txt')
	print 'Initializing FE and Xe'
	FE_Mdb(FM_Name)
	#Xe=Xei*np.ones(Ele_NUM,dtype='f8');
	#Xe=np.array(0.001+0.999*np.random.rand(3600),dtype='f8')
	f=open('Xe_Initial.txt')
	Xe=pickle.load(f)
	f.close()
	low=np.zeros(Ele_NUM,dtype='f8')
	upp=np.zeros(Ele_NUM,dtype='f8')
else:
	# Load Xe low and upp
	FE_Mdb(FM_Name)
	print 'Loading Optimization from: Iter  '+str(iter_num)
	f=open('Xe_'+str(iter_num)+'.txt','r')
	Xe=pickle.load(f)
	low=pickle.load(f)
	upp=pickle.load(f)
	f.close()

print 'Loading......'+FM_Name
f=open(FM_Name,'r')
Fm=pickle.load(f)
f.close()
# Start Time
to=time.time()
# Run Topology optimization
Topot_Main(iter_num+1,Xe,upp,low,change,iter_num)